/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>

#include "xroar.h"
#include "m6809.h"
#include "sam.h"
#include "pia.h"
#include "keyboard.h"
#include "sound.h"
#include "logging.h"
#include "snapshot.h"
#include "tape.h"

/* Condition Code manipulation macros */

#define CC_E 0x80
#define CC_F 0x40
#define CC_H 0x20
#define CC_I 0x10
#define CC_N 0x08
#define CC_Z 0x04
#define CC_V 0x02
#define CC_C 0x01
#define N_EOR_V	((reg_cc & CC_N)^((reg_cc & CC_V)<<2))

#define CLR_HNZVC	reg_cc &= ~(CC_H|CC_N|CC_Z|CC_V|CC_C)
#define CLR_NZV		reg_cc &= ~(CC_N|CC_Z|CC_V)
#define CLR_NZVC	reg_cc &= ~(CC_N|CC_Z|CC_V|CC_C)
#define CLR_Z		reg_cc &= ~(CC_Z)
#define CLR_NZC		reg_cc &= ~(CC_N|CC_Z|CC_C)
#define CLR_NVC		reg_cc &= ~(CC_N|CC_V|CC_C)
#define CLR_ZC		reg_cc &= ~(CC_Z|CC_C)

#define SET_Z(a)	if (!(a)) reg_cc |= CC_Z
#define SET_N8(a)	reg_cc |= (a&0x80)>>4
#define SET_N16(a)	reg_cc |= (a&0x8000)>>12
#define SET_H(a,b,r)	reg_cc |= ((a^b^r)&0x10)<<1
#define SET_C8(a)	reg_cc |= (a&0x100)>>8
#define SET_C16(a)	reg_cc |= (a&0x10000)>>16
#define SET_V8(a,b,r)	reg_cc |= ((a^b^r^(r>>1))&0x80)>>6
#define SET_V16(a,b,r)  reg_cc |= ((a^b^r^(r>>1))&0x8000)>>14
#define SET_NZ8(a)		SET_N8(a);SET_Z(a&0xff)
#define SET_NZ16(a)		SET_N16(a);SET_Z(a&0xffff)
#define SET_NZC8(a)		SET_N8(a);SET_Z(a&0xff);SET_C8(a)
#define SET_NZVC8(a,b,r)	SET_N8(r);SET_Z(r&0xff);SET_V8(a,b,r);SET_C8(r)
#define SET_NZVC16(a,b,r)	SET_N16(r);SET_Z(r&0xffff);SET_V16(a,b,r);SET_C16(r)

#define EAD_DIRECT	EAD = reg_dp << 8 | IMMEDIATE_BYTE
#define EAD_INDEXED	addr_indexed()
#define EAD_EXTENDED	EAD = IMMEDIATE_WORD
#define SHORT_RELATIVE	RA = sex(IMMEDIATE_BYTE)
#define LONG_RELATIVE	RA = sex16(IMMEDIATE_WORD)
#ifdef TRACE
#define IMMEDIATE_BYTE	immbyte()
#else
#define IMMEDIATE_BYTE	sam_read_byte(reg_pc++)
#endif
#define IMMEDIATE_WORD	(IMMEDIATE_BYTE << 8 | IMMEDIATE_BYTE)
#define DIRECT_BYTE	sam_read_byte(EAD_DIRECT)
#define DIRECT_WORD	sam_read_byte(EAD_DIRECT) << 8 | sam_read_byte(EAD+1)
#define EXTENDED_BYTE	sam_read_byte(EAD_EXTENDED)
#define EXTENDED_WORD	sam_read_byte(EAD_EXTENDED) << 8 | sam_read_byte(EAD+1)
#define INDEXED_BYTE	sam_read_byte(EAD_INDEXED)
#define INDEXED_WORD	sam_read_byte(EAD_INDEXED) << 8 | sam_read_byte(EAD+1)
#define STORE_BYTE(o)	sam_store_byte(EAD, o)
#define STORE_WORD(o)	sam_store_byte(EAD, o >> 8); sam_store_byte(EAD+1, o & 0xff)

#define PUSHWORD(s,r)	sam_store_byte(--s, r & 0xff); sam_store_byte(--s, r >> 8);
#define PUSHBYTE(s,r)	sam_store_byte(--s, r);
#define PULLWORD(s,r)	r = sam_read_byte(s++) << 8 | sam_read_byte(s++);
#define PULLBYTE(s,r)	r = sam_read_byte(s++);

Uint32 current_cycle;
int runcycles = 0;

typedef Uint32 (*op_func)(void);
static op_func m6809_page0[];
static op_func m6809_page2[];
static op_func m6809_page3[];

static inline Uint32 sex(Uint32 reg);
static inline Uint32 sex5(Uint32 reg);
static inline Uint32 sex16(Uint32 reg);

/* MPU registers */
Uint8 reg_cc, reg_dp;
Uint16 reg_x, reg_y, reg_u, reg_s, reg_pc;
accumulator_t reg_accum;
#ifdef TRACE
Uint32 immbyte(void) { Uint8 x = sam_read_byte(reg_pc++); if (trace) fprintf(stderr, "%02x ", x); return x; }
#endif

/* MPU work variables */
static Uint32 EAD, RA;
static Uint32 octet;
static Uint32 cycles;  /* used when we need to compute extra cycles */
static Uint32 poll_cycles;  /* periodically poll keyboard and sound */

/* MPU interrupt state variables */
Uint32 nmi, firq, irq;
Uint32 wait_for_interrupt;
Uint32 skip_register_push;

/* ------------------------------------------------------------------------- */

static Uint32 addr_indexed(void) {
	Uint32 postbyte = IMMEDIATE_BYTE;
	Uint16 *word_ptr;
	switch ((postbyte & 0x60)>>5) {
		default:
		case 0: word_ptr = &reg_x; break;
		case 1: word_ptr = &reg_y; break;
		case 2: word_ptr = &reg_u; break;
		case 3: word_ptr = &reg_s; break;
	}
	if (!(postbyte & 0x80)) {
		EAD = *word_ptr + sex5(postbyte);
		cycles = 2;
		return EAD;
	}
	switch (postbyte & 0x0f) {
		case 0x00:  /* ,R+ */
			EAD = *word_ptr;
			*word_ptr += 1;
			cycles = 2;
			break;
		case 0x01:  /* ,R++ */
			EAD = *word_ptr;
			*word_ptr += 2;
			cycles = 3;
			break;
		case 0x02:  /* ,-R */
			*word_ptr -= 1;
			EAD = *word_ptr;
			cycles = 2;
			break;
		case 0x03:  /* ,--R */
			*word_ptr -= 2;
			EAD = *word_ptr;
			cycles = 3;
			break;
		case 0x04:  /* ,R */
			EAD = *word_ptr;
			cycles = 0;
			break;
		case 0x05:  /* B,R */
			EAD = *word_ptr + sex(reg_b);
			cycles = 1;
			break;
		case 0x06:  /* A,R */
			EAD = *word_ptr + sex(reg_a);
			cycles = 1;
			break;
		case 0x08:  /* 8n,R */
			EAD = *word_ptr + sex(IMMEDIATE_BYTE);
			cycles = 1;
			break;
		case 0x09:  /* 16n, R */
			EAD = *word_ptr + sex16(IMMEDIATE_WORD);
			cycles = 4;
			break;
		case 0x0b:  /* D,R */
			EAD = *word_ptr + reg_d;
			cycles = 4;
			break;
		case 0x0c:  /* 8n,PCR */
			EAD = sex(IMMEDIATE_BYTE) + reg_pc;
			cycles = 1;
			break;
		case 0x0d:  /* 16n,PCR */
			EAD = sex16(IMMEDIATE_WORD) + reg_pc;
			cycles = 5;
			break;
		case 0x0f:
			EAD = IMMEDIATE_WORD;
			cycles = 2;
			break;
		default:
			break;
	}
	if (postbyte & 0x10) {
		EAD = sam_read_byte(EAD) << 8 | sam_read_byte(EAD+1);
		cycles += 3;
	}
	cycles++;
	return EAD;
}

/* ------------------------------------------------------------------------- */

static inline Uint32 sex(Uint32 reg) {
	return ((reg&0x80)?0xffffff00:0x00000000)|(reg&0xff);
}
static inline Uint32 sex5(Uint32 reg) {
	return ((reg&0x10)?0xfffffff0:0x00000000)|(reg&0x0f);
}
static inline Uint32 sex16(Uint32 reg) {
	return ((reg&0x8000)?0xffff0000:0x00000000)|(reg&0xffff);
}

/* ------------------------------------------------------------------------- */

inline void m6809_nmi(void) { nmi = 1; }
inline void m6809_firq(void) { firq = 1; }
inline void m6809_irq(void) { irq = 1; }

void m6809_init(void) {
	current_cycle = 0;
	poll_cycles = 0;
	/* PIA_0B.intfunc = m6809_irq;
	PIA_1B.intfunc = m6809_firq; */
}

void m6809_reset(void) {
	reg_cc = 0xff;
	reg_dp = reg_d = reg_x = reg_y = reg_u = reg_s = 0;
	nmi = firq = irq = 0;
	wait_for_interrupt = 0;
	skip_register_push = 0;
	reg_pc = sam_read_word(0xfffe);
}

/* Note:  in the following if (e) and if (!e) should get optimised away.
 * Hopefully that's also the case where m is 0 (ie, NMI) */
#define TEST_INTERRUPT(i,m,v,e,cm,cyc) if (i) { \
	wait_for_interrupt = 0; \
	i = 0; \
	if (!(reg_cc & m)) { \
		if (!skip_register_push) { \
			if (e) reg_cc |= CC_E; if (!e) reg_cc &= ~CC_E; \
			PUSHWORD(reg_s, reg_pc); \
			if (e) { \
				PUSHWORD(reg_s, reg_u); \
				PUSHWORD(reg_s, reg_y); \
				PUSHWORD(reg_s, reg_x); \
				PUSHBYTE(reg_s, reg_dp); \
				PUSHBYTE(reg_s, reg_b); \
				PUSHBYTE(reg_s, reg_a); \
			} \
			PUSHBYTE(reg_s, reg_cc); \
		} \
		skip_register_push = 0; \
		reg_cc |= cm; reg_pc = sam_read_word(v); \
		runcycles -= cyc; current_cycle += cyc; \
	} \
}

void m6809_cycle(void) {
	Uint32 op_ran;
	runcycles += 57;
	TEST_INTERRUPT(nmi, 0, 0xfffc, CC_E, CC_F|CC_I, 18);
	TEST_INTERRUPT(firq, CC_F, 0xfff6, 0, CC_F|CC_I, 9);
	TEST_INTERRUPT(irq, CC_I, 0xfff8, CC_E, CC_I, 18);
	while (runcycles > 0) {
		/* Temporary hack while HSYNC IRQ unimplemented 
		 * (at least I *think* that's what's slowing everything
		 * down) */
		//if (wait_for_interrupt && !PIA_0A.interrupt_enable) {
		if (wait_for_interrupt) {
			current_cycle += runcycles;
			poll_cycles += runcycles;
			runcycles = 0;
		} else {
			/* Trap cassette ROM routines - hack! */
			if (reg_pc == 0xbde7)
				m6809_page0[0x39]();
			if (reg_pc == 0xbda5) {
				tape_read_bit();
				m6809_page0[0x39]();
			}
			if (trace) fprintf(stderr, "%04x| ", reg_pc);
			op_ran = m6809_page0[IMMEDIATE_BYTE]();
			if (trace) fprintf(stderr, "\tcc=%02x a=%02x b=%02x dp=%02x x=%04x y=%04x u=%04x s=%04x\n", reg_cc, reg_a, reg_b, reg_dp, reg_x, reg_y, reg_u, reg_s);
			runcycles -= op_ran;
			current_cycle += op_ran;
			poll_cycles += op_ran;
		}
		while (poll_cycles > 10) {
			sound_write_byte(PIA_1A.output_register);
			poll_cycles -= 10;
		}
	}
	return;
}

/* ------------------------------------------------------------------------- */

/* Operation templates */

#define OP_NEG(a,c) Uint32 result; octet = a; result = ~(octet) + 1; CLR_NZVC; SET_NZVC8(0, octet, result); STORE_BYTE(result); return c
#define OP_COM(a,c) octet = a; octet = ~(octet); CLR_NZV; SET_NZ8(octet); reg_cc |= CC_C; STORE_BYTE(octet); return c
#define OP_LSR(a,c) octet = a; CLR_NZC; reg_cc |= (octet & CC_C); octet >>= 1; SET_Z(octet); STORE_BYTE(octet); return c
#define OP_ROR(a,c) Uint32 result; octet = a; result = (reg_cc & CC_C) << 7; CLR_NZC; reg_cc |= octet & CC_C; result |= octet >> 1; SET_NZ8(result); STORE_BYTE(result); return c
#define OP_ASR(a,c) CLR_NZC; octet = a; reg_cc |= (octet & CC_C); octet = (octet & 0x80) | (octet >> 1); SET_NZ8(octet); STORE_BYTE(octet); return c
#define OP_ASL(a,c) Uint32 result; octet = a; result = octet << 1; CLR_NZVC; SET_NZVC8(octet, octet, result); STORE_BYTE(result); return c
#define OP_ROL(a,c) Uint32 result; octet = a; result = (octet<<1)|(reg_cc & CC_C); CLR_NZVC; SET_NZVC8(octet, octet, result); STORE_BYTE(result); return c
#define OP_DEC(a,c) octet = a; octet--; CLR_NZV; SET_NZ8(octet); if (octet == 0x7f) reg_cc |= CC_V; STORE_BYTE(octet); return c
#define OP_INC(a,c) octet = a; octet++; CLR_NZV; SET_NZ8(octet); if (octet == 0x80) reg_cc |= CC_V; STORE_BYTE(octet); return c
#define OP_TST(a,c) octet = a; CLR_NZV; SET_NZ8(octet); return c
#define OP_JMP(a,c) a; reg_pc = EAD; return c
#define OP_CLR(a,c) octet = a; STORE_BYTE(0); CLR_NVC; reg_cc |= CC_Z; return c

#define OP_NEGR(r,c) Uint32 result = ~(r) + 1; CLR_NZVC; SET_NZVC8(0, r, result); r = result; return c
#define OP_COMR(r,c) r = ~(r); CLR_NZV; SET_NZ8(r); reg_cc |= CC_C; return c
#define OP_LSRR(r,c) CLR_NZC; reg_cc |= (r & CC_C); r >>= 1; SET_Z(r); return c
#define OP_RORR(r,c) Uint32 result = (reg_cc & CC_C) << 7; CLR_NZC; reg_cc |= r & CC_C; result |= r >> 1; SET_NZ8(result); r = result; return c
#define OP_ASRR(r,c) CLR_NZC; reg_cc |= (r & CC_C); r = (r & 0x80) | (r >> 1); SET_NZ8(r); return c
#define OP_ASLR(r,c) Uint32 result = r << 1; CLR_NZVC; SET_NZVC8(r, r, result); r = result; return c
#define OP_ROLR(r,c) Uint32 result = (reg_cc & CC_C) | (r << 1); CLR_NZVC; SET_NZVC8(r, r, result); r = result; return c
#define OP_DECR(r,c) r--; CLR_NZV; SET_NZ8(r); if (r == 0x7f) reg_cc |= CC_V; return c
#define OP_INCR(r,c) r++; CLR_NZV; SET_NZ8(r); if (r == 0x80) reg_cc |= CC_V; return c
#define OP_TSTR(r,c) CLR_NZV; SET_NZ8(r); return c
#define OP_CLRR(r,c) r = 0; CLR_NVC; reg_cc |= CC_Z; return c

#define OP_SUB16(r,a,c) Uint32 result; octet = a; result = r - octet; CLR_NZVC; SET_NZVC16(r, octet, result); r = result; return c
#define OP_ADD16(r,a,c) Uint32 result; octet = a; result = r + octet; CLR_NZVC; SET_NZVC16(r, octet, result); r = result; return c
#define OP_CMP16(r,a,c) Uint32 result; octet = a; result = r - octet; CLR_NZVC; SET_NZVC16(r, octet, result); return c
#define OP_LD16(r,a,c) r = a; CLR_NZV; SET_NZ16(r); return c
#define OP_ST16(r,a,c) a; sam_store_byte(EAD, r >> 8); sam_store_byte(EAD+1, r & 0xff); CLR_NZV; SET_NZ16(r); return c
#define OP_JSR(a,c) Uint32 dest = a; PUSHWORD(reg_s, reg_pc); reg_pc = dest; return c

#define OP_SUB(r,a,c) Uint32 result; octet = a; result = r - octet; CLR_NZVC; SET_NZVC8(r, octet, result); r = result; return c
#define OP_CMP(r,a,c) Uint32 result; octet = a; result = r - octet; CLR_NZVC; SET_NZVC8(r, octet, result); return c
#define OP_SBC(r,a,c) Uint32 result; octet = a; result = r - octet - (reg_cc & CC_C); CLR_NZVC; SET_NZVC8(r, octet, result); r = result; return c
#define OP_AND(r,a,c) r &= a; CLR_NZV; SET_NZ8(r); return c
#define OP_BIT(r,a,c) Uint32 result = r & a; CLR_NZV; SET_NZ8(result); return c
#define OP_LD(r,a,c) r = a; CLR_NZV; SET_NZ8(r); return c
#define OP_ST(r,a,c) a; sam_store_byte(EAD, r); CLR_NZV; SET_NZ8(r); return c
#define OP_EOR(r,a,c) r ^= a; CLR_NZV; SET_NZ8(r); return c
#define OP_ADC(r,a,c) Uint32 result; octet = a; result = r + octet + (reg_cc & CC_C); CLR_HNZVC; SET_NZVC8(r, octet, result); SET_H(r, octet, result); r = result; return c
#define OP_OR(r,a,c) r |= a; CLR_NZV; SET_NZ8(r); return c
#define OP_ADD(r,a,c) Uint32 result; octet = a; result = r + octet; CLR_HNZVC; SET_NZVC8(r, octet, result); SET_H(r, octet, result); r = result; return c

/* ------------------------------------------------------------------------- */

/* 0x00 NEG direct */
static Uint32 op_neg_direct(void) { OP_NEG(DIRECT_BYTE, 6); }
/* 0x03 COM direct */
static Uint32 op_com_direct(void) { OP_COM(DIRECT_BYTE, 6); }
/* 0x01 NEGCOM direct (illegal) */
static Uint32 op_negcom_direct(void) {
	if (reg_cc & CC_C)
		return op_com_direct();
	return op_neg_direct();
}
/* 0x04 LSR direct */
static Uint32 op_lsr_direct(void) { OP_LSR(DIRECT_BYTE, 6); }
/* 0x06 ROR direct */
static Uint32 op_ror_direct(void) { OP_ROR(DIRECT_BYTE, 6); }
/* 0x07 ASR direct */
static Uint32 op_asr_direct(void) { OP_ASR(DIRECT_BYTE, 6); }
/* 0x08 ASL,LSL direct */
static Uint32 op_asl_direct(void) { OP_ASL(DIRECT_BYTE, 6); }
/* 0x09 ROL direct */
static Uint32 op_rol_direct(void) { OP_ROL(DIRECT_BYTE, 6); }
/* 0x0A DEC direct */
static Uint32 op_dec_direct(void) { OP_DEC(DIRECT_BYTE, 6); }
/* 0x0C INC direct */
static Uint32 op_inc_direct(void) { OP_INC(DIRECT_BYTE, 6); }
/* 0x0D TST direct */
static Uint32 op_tst_direct(void) { OP_TST(DIRECT_BYTE, 6); }
/* 0x0E JMP direct */
static Uint32 op_jmp_direct(void) { OP_JMP(EAD_DIRECT,  3); }
/* 0x0F CLR direct */
static Uint32 op_clr_direct(void) { OP_CLR(DIRECT_BYTE, 6); }

/* ------------------------------------------------------------------------- */

/* 0x10 Page 2 */
static Uint32 op_page2(void) {
	return m6809_page2[IMMEDIATE_BYTE]();
}

/* 0x11 Page 3 */
static Uint32 op_page3(void) {
	return m6809_page3[IMMEDIATE_BYTE]();
}

/* 0x12 NOP inherent */
static Uint32 op_nop(void) {
	return 2;
}

/* 0x13 SYNC inherent */
static Uint32 op_sync(void) {
	wait_for_interrupt = 1;
	return 3;
}

/* 0x16 LBRA relative */
static Uint32 op_lbra(void) {
	LONG_RELATIVE;
	reg_pc += RA;
	return 5;
}

/* 0x17 LBSR relative */
static Uint32 op_lbsr(void) {
	LONG_RELATIVE;
	PUSHWORD(reg_s, reg_pc);
	reg_pc += RA;
	return 9;
}

/* 0x19 DAA inherent */
static Uint32 op_daa(void) {
	Uint32 result = 0;
	if ((reg_a&0x0f) >= 0x0a || reg_cc & CC_H) result |= 0x06;
	if (reg_a >= 0x90 && (reg_a&0x0f) >= 0x0a) result |= 0x60;
	if (reg_a >= 0xa0 || reg_cc & CC_C) result |= 0x60;
	result += reg_a;
	reg_a = result & 0xff;
	CLR_NZV;
	SET_NZC8(result);
	return 2;
}

/* 0x1A ORCC immediate */
static Uint32 op_orcc(void) {
	reg_cc |= IMMEDIATE_BYTE;
	return 3;
}

/* 0x1C ANDCC immediate */
static Uint32 op_andcc(void) {
	reg_cc &= IMMEDIATE_BYTE;
	return 3;
}

/* 0x1D SEX inherent */
static Uint32 op_sex(void) {
	reg_d = sex(reg_a);
	CLR_NZV;
	SET_NZ16(reg_d);
	return 2;
}

/* 0x1E EXG immediate */
static Uint32 op_exg(void) {
	int postbyte = IMMEDIATE_BYTE;
	Uint8 s_octet = 0xff, d_octet = 0xff, tmp_octet;
	Uint16 s_word = 0xffff, d_word = 0xffff, tmp_word;
	Uint8 *source_octet = &s_octet, *dest_octet = &d_octet;
	Uint16 *source_word = &s_word, *dest_word = &d_word;
	switch (postbyte & 0xf0) {
		case 0x00: source_word = &reg_d; break;
		case 0x10: source_word = &reg_x; break;
		case 0x20: source_word = &reg_y; break;
		case 0x30: source_word = &reg_u; break;
		case 0x40: source_word = &reg_s; break;
		case 0x50: source_word = &reg_pc; break;
		case 0x80: source_octet = &reg_a; break;
		case 0x90: source_octet = &reg_b; break;
		case 0xa0: source_octet = &reg_cc; break;
		case 0xb0: source_octet = &reg_dp; break;
		default: break;
	}
	s_octet = *source_word & 0xff;
	d_word = 0xff00 | *source_octet;
	switch (postbyte & 0x0f) {
		case 0x00: dest_word = &reg_d; break;
		case 0x01: dest_word = &reg_x; break;
		case 0x02: dest_word = &reg_y; break;
		case 0x03: dest_word = &reg_u; break;
		case 0x04: dest_word = &reg_s; break;
		case 0x05: dest_word = &reg_pc; break;
		case 0x08: dest_octet = &reg_a; break;
		case 0x09: dest_octet = &reg_b; break;
		case 0x0a: dest_octet = &reg_cc; break;
		case 0x0b: dest_octet = &reg_dp; break;
		default: break;
	}
	d_octet = *dest_word & 0xff;
	d_word = 0xff00 | *dest_octet;
	tmp_octet = *dest_octet;
	tmp_word = *dest_word;
	*dest_octet = *source_octet;
	*dest_word = *source_word;
	*source_octet = tmp_octet;
	*source_word = tmp_word;
	return 8;
}

/* 0x1F TFR immediate */
static Uint32 op_tfr(void) {
	int postbyte = IMMEDIATE_BYTE;
	Uint8 s_octet = 0xff;
	Uint16 s_word = 0xffff;
	Uint8 *source_octet = &s_octet, *dest_octet = &s_octet;
	Uint16 *source_word = &s_word, *dest_word = &s_word;
	switch (postbyte & 0xf0) {
		case 0x00: source_word = &reg_d; break;
		case 0x10: source_word = &reg_x; break;
		case 0x20: source_word = &reg_y; break;
		case 0x30: source_word = &reg_u; break;
		case 0x40: source_word = &reg_s; break;
		case 0x50: source_word = &reg_pc; break;
		case 0x80: source_octet = &reg_a; break;
		case 0x90: source_octet = &reg_b; break;
		case 0xa0: source_octet = &reg_cc; break;
		case 0xb0: source_octet = &reg_dp; break;
		default: break;
	}
	s_octet = *source_word & 0xff;
	s_word = 0xff00 | *source_octet;
	switch (postbyte & 0x0f) {
		case 0x00: dest_word = &reg_d; break;
		case 0x01: dest_word = &reg_x; break;
		case 0x02: dest_word = &reg_y; break;
		case 0x03: dest_word = &reg_u; break;
		case 0x04: dest_word = &reg_s; break;
		case 0x05: dest_word = &reg_pc; break;
		case 0x08: dest_octet = &reg_a; break;
		case 0x09: dest_octet = &reg_b; break;
		case 0x0a: dest_octet = &reg_cc; break;
		case 0x0b: dest_octet = &reg_dp; break;
		default: break;
	}
	*dest_octet = *source_octet;
	*dest_word = *source_word;
	return 6;
}

/* ------------------------------------------------------------------------- */

/* 0x20 BRA relative */
static Uint32 op_bra(void) {
	SHORT_RELATIVE;
	reg_pc += RA;
	return 3;
}

/* 0x21 BRN relative */
static Uint32 op_brn(void) {
	SHORT_RELATIVE;
	return 3;
}

/* 0x22 BHI relative */
static Uint32 op_bhi(void) {
	SHORT_RELATIVE;
	if (!(reg_cc & (CC_Z|CC_C)))
		reg_pc += RA;
	return 3;
}

/* 0x23 BLS relative */
static Uint32 op_bls(void) {
	SHORT_RELATIVE;
	if (reg_cc & (CC_Z|CC_C))
		reg_pc += RA;
	return 3;
}

/* 0x24 BHS,BCC relative */
static Uint32 op_bhs(void) {
	SHORT_RELATIVE;
	if (!(reg_cc & CC_C))
		reg_pc += RA;
	return 3;
}

/* 0x25 BLO,BCS relative */
static Uint32 op_blo(void) {
	SHORT_RELATIVE;
	if (reg_cc & CC_C)
		reg_pc += RA;
	return 3;
}

/* 0x26 BNE relative */
static Uint32 op_bne(void) {
	SHORT_RELATIVE;
	if (!(reg_cc & CC_Z))
		reg_pc += RA;
	return 3;
}

/* 0x27 BEQ relative */
static Uint32 op_beq(void) {
	SHORT_RELATIVE;
	if (reg_cc & CC_Z)
		reg_pc += RA;
	return 3;
}

/* 0x28 BVC relative */
static Uint32 op_bvc(void) {
	SHORT_RELATIVE;
	if (!(reg_cc & CC_V))
		reg_pc += RA;
	return 3;
}

/* 0x29 BVS relative */
static Uint32 op_bvs(void) {
	SHORT_RELATIVE;
	if (reg_cc & CC_V)
		reg_pc += RA;
	return 3;
}

/* 0x2A BPL relative */
static Uint32 op_bpl(void) {
	SHORT_RELATIVE;
	if (!(reg_cc & CC_N))
		reg_pc += RA;
	return 3;
}

/* 0x2B BMI relative */
static Uint32 op_bmi(void) {
	SHORT_RELATIVE;
	if (reg_cc & CC_N)
		reg_pc += RA;
	return 3;
}

/* 0x2C BGE relative */
static Uint32 op_bge(void) {
	SHORT_RELATIVE;
	if (!N_EOR_V)
		reg_pc += RA;
	return 3;
}

/* 0x2D BLT relative */
static Uint32 op_blt(void) {
	SHORT_RELATIVE;
	if (N_EOR_V)
		reg_pc += RA;
	return 3;
}

/* 0x2E BGT relative */
static Uint32 op_bgt(void) {
	SHORT_RELATIVE;
	if (!(N_EOR_V || reg_cc & CC_Z))
		reg_pc += RA;
	return 3;
}

/* 0x2F BLE relative */
static Uint32 op_ble(void) {
	SHORT_RELATIVE;
	if (N_EOR_V || reg_cc & CC_Z)
		reg_pc += RA;
	return 3;
}

/* ------------------------------------------------------------------------- */

/* 0x30 LEAX indexed */

static Uint32 op_leax(void) {
	EAD_INDEXED;
	reg_x = EAD;
	CLR_Z;
	SET_Z(reg_x);
	return cycles + 4;
}

/* 0x31 LEAY indexed */

static Uint32 op_leay(void) {
	EAD_INDEXED;
	reg_y = EAD;
	CLR_Z;
	SET_Z(reg_y);
	return cycles + 4;
}

/* 0x32 LEAS indexed */

static Uint32 op_leas(void) {
	EAD_INDEXED;
	reg_s = EAD;
	return cycles + 4;
}

/* 0x33 LEAU indexed */

static Uint32 op_leau(void) {
	EAD_INDEXED;
	reg_u = EAD;
	return cycles + 4;
}

/* 0x34 PSHS immediate */

static Uint32 op_pshs(void) {
	int postbyte = IMMEDIATE_BYTE;
	cycles = 0;
	if (postbyte & 0x80) { PUSHWORD(reg_s, reg_pc); cycles += 2; }
	if (postbyte & 0x40) { PUSHWORD(reg_s, reg_u); cycles += 2; }
	if (postbyte & 0x20) { PUSHWORD(reg_s, reg_y); cycles += 2; }
	if (postbyte & 0x10) { PUSHWORD(reg_s, reg_x); cycles += 2; }
	if (postbyte & 0x08) { PUSHBYTE(reg_s, reg_dp); cycles++; }
	if (postbyte & 0x04) { PUSHBYTE(reg_s, reg_b); cycles++; }
	if (postbyte & 0x02) { PUSHBYTE(reg_s, reg_a); cycles++; }
	if (postbyte & 0x01) { PUSHBYTE(reg_s, reg_cc); cycles++; }
	return cycles + 5;
}

/* 0x35 PULS immediate */

static Uint32 op_puls(void) {
	int postbyte = IMMEDIATE_BYTE;
	cycles = 0;
	if (postbyte & 0x01) { PULLBYTE(reg_s, reg_cc); cycles++; }
	if (postbyte & 0x02) { PULLBYTE(reg_s, reg_a); cycles++; }
	if (postbyte & 0x04) { PULLBYTE(reg_s, reg_b); cycles++; }
	if (postbyte & 0x08) { PULLBYTE(reg_s, reg_dp); cycles++; }
	if (postbyte & 0x10) { PULLWORD(reg_s, reg_x); cycles += 2; }
	if (postbyte & 0x20) { PULLWORD(reg_s, reg_y); cycles += 2; }
	if (postbyte & 0x40) { PULLWORD(reg_s, reg_u); cycles += 2; }
	if (postbyte & 0x80) { PULLWORD(reg_s, reg_pc); cycles += 2; }
	return cycles + 5;
}

/* 0x36 PSHU immediate */

static Uint32 op_pshu(void) {
	int postbyte = IMMEDIATE_BYTE;
	cycles = 0;
	if (postbyte & 0x80) { PUSHWORD(reg_u, reg_pc); cycles += 2; }
	if (postbyte & 0x40) { PUSHWORD(reg_u, reg_s); cycles += 2; }
	if (postbyte & 0x20) { PUSHWORD(reg_u, reg_y); cycles += 2; }
	if (postbyte & 0x10) { PUSHWORD(reg_u, reg_x); cycles += 2; }
	if (postbyte & 0x08) { PUSHBYTE(reg_u, reg_dp); cycles++; }
	if (postbyte & 0x04) { PUSHBYTE(reg_u, reg_b); cycles++; }
	if (postbyte & 0x02) { PUSHBYTE(reg_u, reg_a); cycles++; }
	if (postbyte & 0x01) { PUSHBYTE(reg_u, reg_cc); cycles++; }
	return cycles + 5;
}

/* 0x37 PULU immediate */

static Uint32 op_pulu(void) {
	int postbyte = IMMEDIATE_BYTE;
	cycles = 0;
	if (postbyte & 0x01) { PULLBYTE(reg_u, reg_cc); cycles++; }
	if (postbyte & 0x02) { PULLBYTE(reg_u, reg_a); cycles++; }
	if (postbyte & 0x04) { PULLBYTE(reg_u, reg_b); cycles++; }
	if (postbyte & 0x08) { PULLBYTE(reg_u, reg_dp); cycles++; }
	if (postbyte & 0x10) { PULLWORD(reg_u, reg_x); cycles += 2; }
	if (postbyte & 0x20) { PULLWORD(reg_u, reg_y); cycles += 2; }
	if (postbyte & 0x40) { PULLWORD(reg_u, reg_s); cycles += 2; }
	if (postbyte & 0x80) { PULLWORD(reg_u, reg_pc); cycles += 2; }
	return cycles + 5;
}

/* 0x39 RTS inherent */

static Uint32 op_rts(void) {
	PULLWORD(reg_s, reg_pc);
	return 5;
}

/* 0x3A ABX inherent */

static Uint32 op_abx(void) {
	reg_x += sex(reg_b);
	return 3;
}

/* 0x3B RTI inherent */

static Uint32 op_rti(void) {
	PULLBYTE(reg_s, reg_cc);
	if (reg_cc & CC_E) {
		PULLBYTE(reg_s, reg_a);
		PULLBYTE(reg_s, reg_b);
		PULLBYTE(reg_s, reg_dp);
		PULLWORD(reg_s, reg_x);
		PULLWORD(reg_s, reg_y);
		PULLWORD(reg_s, reg_u);
		PULLWORD(reg_s, reg_pc);
		return 15;
	} else {
		PULLWORD(reg_s, reg_pc);
		return 6;
	}
}

/* 0x3C CWAI immediate */

static Uint32 op_cwai(void) {
	reg_cc |= CC_E;
	PUSHWORD(reg_s, reg_pc);
	PUSHWORD(reg_s, reg_u);
	PUSHWORD(reg_s, reg_y);
	PUSHWORD(reg_s, reg_x);
	PUSHBYTE(reg_s, reg_dp);
	PUSHBYTE(reg_s, reg_b);
	PUSHBYTE(reg_s, reg_a);
	PUSHBYTE(reg_s, reg_cc);
	wait_for_interrupt = 1;
	skip_register_push = 1;
	return 11;
}

/* 0x3D MUL inherent */
static Uint32 op_mul(void) {
	reg_d = reg_a * reg_b;
	CLR_ZC;
	SET_Z(reg_d);
	if (reg_d & 0x80)
		reg_cc |= CC_C;
	return 11;
}

/* 0x3F SWI inherent */
static Uint32 op_swi(void) {
	reg_cc |= CC_E;
	PUSHWORD(reg_s, reg_pc);
	PUSHWORD(reg_s, reg_u);
	PUSHWORD(reg_s, reg_y);
	PUSHWORD(reg_s, reg_x);
	PUSHBYTE(reg_s, reg_dp);
	PUSHBYTE(reg_s, reg_b);
	PUSHBYTE(reg_s, reg_a);
	PUSHBYTE(reg_s, reg_cc);
	reg_cc |= CC_F|CC_I;
	reg_pc = sam_read_word(0xfffa);
	return 19;
}

/* ------------------------------------------------------------------------- */

/* 0x40 NEGA inherent */
static Uint32 op_nega(void) { OP_NEGR(reg_a, 6); }
/* 0x43 COMA inherent */
static Uint32 op_coma(void) { OP_COMR(reg_a, 6); }
/* 0x44 LSRA inherent */
static Uint32 op_lsra(void) { OP_LSRR(reg_a, 6); }
/* 0x46 RORA inherent */
static Uint32 op_rora(void) { OP_RORR(reg_a, 6); }
/* 0x47 ASRA inherent */
static Uint32 op_asra(void) { OP_ASRR(reg_a, 6); }
/* 0x48 ASLA,LSLA inherent */
static Uint32 op_asla(void) { OP_ASLR(reg_a, 6); }
/* 0x49 ROLA inherent */
static Uint32 op_rola(void) { OP_ROLR(reg_a, 6); }
/* 0x4A DECA inherent */
static Uint32 op_deca(void) { OP_DECR(reg_a, 6); }
/* 0x4C INCA inherent */
static Uint32 op_inca(void) { OP_INCR(reg_a, 6); }
/* 0x4D TSTA inherent */
static Uint32 op_tsta(void) { OP_TSTR(reg_a, 6); }
/* 0x4F CLRA inherent */
static Uint32 op_clra(void) { OP_CLRR(reg_a, 6); }

/* ------------------------------------------------------------------------- */

/* 0x50 NEGB inherent */
static Uint32 op_negb(void) { OP_NEGR(reg_b, 6); }
/* 0x53 COMB inherent */
static Uint32 op_comb(void) { OP_COMR(reg_b, 6); }
/* 0x54 LSRB inherent */
static Uint32 op_lsrb(void) { OP_LSRR(reg_b, 6); }
/* 0x56 RORB inherent */
static Uint32 op_rorb(void) { OP_RORR(reg_b, 6); }
/* 0x57 ASRB inherent */
static Uint32 op_asrb(void) { OP_ASRR(reg_b, 6); }
/* 0x58 ASLB,LSLB inherent */
static Uint32 op_aslb(void) { OP_ASLR(reg_b, 6); }
/* 0x59 ROLB inherent */
static Uint32 op_rolb(void) { OP_ROLR(reg_b, 6); }
/* 0x5A DECB inherent */
static Uint32 op_decb(void) { OP_DECR(reg_b, 6); }
/* 0x5C INCB inherent */
static Uint32 op_incb(void) { OP_INCR(reg_b, 6); }
/* 0x5D TSTB inherent */
static Uint32 op_tstb(void) { OP_TSTR(reg_b, 6); }
/* 0x5F CLRB inherent */
static Uint32 op_clrb(void) { OP_CLRR(reg_b, 6); }

/* ------------------------------------------------------------------------- */

/* 0x60 NEG indexed */
static Uint32 op_neg_indexed(void) { OP_NEG(INDEXED_BYTE, 6+cycles); }
/* 0x63 COM indexed */
static Uint32 op_com_indexed(void) { OP_COM(INDEXED_BYTE, 6+cycles); }
/* 0x64 LSR indexed */
static Uint32 op_lsr_indexed(void) { OP_LSR(INDEXED_BYTE, 6+cycles); }
/* 0x66 ROR indexed */
static Uint32 op_ror_indexed(void) { OP_ROR(INDEXED_BYTE, 6+cycles); }
/* 0x67 ASR indexed */
static Uint32 op_asr_indexed(void) { OP_ASR(INDEXED_BYTE, 6+cycles); }
/* 0x68 ASL,LSL indexed */
static Uint32 op_asl_indexed(void) { OP_ASL(INDEXED_BYTE, 6+cycles); }
/* 0x69 ROL indexed */
static Uint32 op_rol_indexed(void) { OP_ROL(INDEXED_BYTE, 6+cycles); }
/* 0x6A DEC indexed */
static Uint32 op_dec_indexed(void) { OP_DEC(INDEXED_BYTE, 6+cycles); }
/* 0x6C INC indexed */
static Uint32 op_inc_indexed(void) { OP_INC(INDEXED_BYTE, 6+cycles); }
/* 0x6D TST indexed */
static Uint32 op_tst_indexed(void) { OP_TST(INDEXED_BYTE, 6+cycles); }
/* 0x6E JMP indexed */
static Uint32 op_jmp_indexed(void) { OP_JMP(EAD_INDEXED,  3+cycles); }
/* 0x6F CLR indexed */
static Uint32 op_clr_indexed(void) { OP_CLR(INDEXED_BYTE, 6+cycles); }

/* ------------------------------------------------------------------------- */

/* 0x70 NEG extended */
static Uint32 op_neg_extended(void) { OP_NEG(EXTENDED_BYTE, 7); }
/* 0x73 COM extended */
static Uint32 op_com_extended(void) { OP_COM(EXTENDED_BYTE, 7); }
/* 0x74 LSR extended */
static Uint32 op_lsr_extended(void) { OP_LSR(EXTENDED_BYTE, 7); }
/* 0x76 ROR extended */
static Uint32 op_ror_extended(void) { OP_ROR(EXTENDED_BYTE, 7); }
/* 0x77 ASR extended */
static Uint32 op_asr_extended(void) { OP_ASR(EXTENDED_BYTE, 7); }
/* 0x78 ASL,LSL extended */
static Uint32 op_asl_extended(void) { OP_ASL(EXTENDED_BYTE, 7); }
/* 0x79 ROL extended */
static Uint32 op_rol_extended(void) { OP_ROL(EXTENDED_BYTE, 7); }
/* 0x7A DEC extended */
static Uint32 op_dec_extended(void) { OP_DEC(EXTENDED_BYTE, 7); }
/* 0x7C INC extended */
static Uint32 op_inc_extended(void) { OP_INC(EXTENDED_BYTE, 7); }
/* 0x7D TST extended */
static Uint32 op_tst_extended(void) { OP_TST(EXTENDED_BYTE, 7); }
/* 0x7E JMP extended */
static Uint32 op_jmp_extended(void) { OP_JMP(EAD_EXTENDED,  4); }
/* 0x7F CLR extended */
static Uint32 op_clr_extended(void) { OP_CLR(EXTENDED_BYTE, 7); }

/* ------------------------------------------------------------------------- */

/* 0x80 SUBA immediate */
static Uint32 op_suba(void) { OP_SUB(reg_a, IMMEDIATE_BYTE, 2); }
/* 0x81 CMPA immediate */
static Uint32 op_cmpa(void) { OP_CMP(reg_a, IMMEDIATE_BYTE, 2); }
/* 0x82 SBCA immediate */
static Uint32 op_sbca(void) { OP_SBC(reg_a, IMMEDIATE_BYTE, 2); }
/* 0x83 SUBD immediate */
static Uint32 op_subd(void) { OP_SUB16(reg_d, IMMEDIATE_WORD, 4); }
/* 0x84 ANDA immediate */
static Uint32 op_anda(void) { OP_AND(reg_a, IMMEDIATE_BYTE, 2); }
/* 0x85 BITA immediate */
static Uint32 op_bita(void) { OP_BIT(reg_a, IMMEDIATE_BYTE, 2); }
/* 0x86 LDA immediate */
static Uint32 op_lda(void) { OP_LD(reg_a, IMMEDIATE_BYTE, 2); }
/* 0x88 EORA immediate */
static Uint32 op_eora(void) { OP_EOR(reg_a, IMMEDIATE_BYTE, 2); }
/* 0x89 ADCA immediate */
static Uint32 op_adca(void) { OP_ADC(reg_a, IMMEDIATE_BYTE, 2); }
/* 0x8A ORA immediate */
static Uint32 op_ora(void) { OP_OR(reg_a, IMMEDIATE_BYTE, 2); }
/* 0x8B ADDA immediate */
static Uint32 op_adda(void) { OP_ADD(reg_a, IMMEDIATE_BYTE, 2); }
/* 0x8C CMPX immediate */
static Uint32 op_cmpx(void) { OP_CMP16(reg_x, IMMEDIATE_WORD, 4); }
/* 0x8D BSR relative */
static Uint32 op_bsr(void) {
	SHORT_RELATIVE;
	PUSHWORD(reg_s, reg_pc);
	reg_pc += RA;
	return 7;
}
/* 0x8E LDX immediate */
static Uint32 op_ldx(void) { OP_LD16(reg_x, IMMEDIATE_WORD, 3); }

/* ------------------------------------------------------------------------- */

/* 0x90 SUBA direct */
static Uint32 op_suba_direct(void) { OP_SUB(reg_a, DIRECT_BYTE, 4); }
/* 0x91 CMPA direct */
static Uint32 op_cmpa_direct(void) { OP_CMP(reg_a, DIRECT_BYTE, 4); }
/* 0x92 SBCA direct */
static Uint32 op_sbca_direct(void) { OP_SBC(reg_a, DIRECT_BYTE, 4); }
/* 0x93 SUBD direct */
static Uint32 op_subd_direct(void) { OP_SUB16(reg_d, DIRECT_WORD, 6); }
/* 0x94 ANDA direct */
static Uint32 op_anda_direct(void) { OP_AND(reg_a, DIRECT_BYTE, 4); }
/* 0x95 BITA direct */
static Uint32 op_bita_direct(void) { OP_BIT(reg_a, DIRECT_BYTE, 4); }
/* 0x96 LDA direct */
static Uint32 op_lda_direct(void) { OP_LD(reg_a, DIRECT_BYTE, 4); }
/* 0x97 STA direct */
static Uint32 op_sta_direct(void) { OP_ST(reg_a, EAD_DIRECT, 4); }
/* 0x98 EORA direct */
static Uint32 op_eora_direct(void) { OP_EOR(reg_a, DIRECT_BYTE, 4); }
/* 0x99 ADCA direct */
static Uint32 op_adca_direct(void) { OP_ADC(reg_a, DIRECT_BYTE, 4); }
/* 0x9A ORA direct */
static Uint32 op_ora_direct(void) { OP_OR(reg_a, DIRECT_BYTE, 4); }
/* 0x9B ADDA direct */
static Uint32 op_adda_direct(void) { OP_ADD(reg_a, DIRECT_BYTE, 4); }
/* 0x9C CMPX direct */
static Uint32 op_cmpx_direct(void) { OP_CMP16(reg_x, DIRECT_WORD, 6); }
/* 0x9D JSR direct */
static Uint32 op_jsr_direct(void) { OP_JSR(EAD_DIRECT, 7); }
/* 0x9E LDX direct */
static Uint32 op_ldx_direct(void) { OP_LD16(reg_x, DIRECT_WORD, 5); }
/* 0x9F STX direct */
static Uint32 op_stx_direct(void) { OP_ST16(reg_x, EAD_DIRECT, 5); }

/* ------------------------------------------------------------------------- */

/* 0xA0 SUBA indexed */
static Uint32 op_suba_indexed(void) { OP_SUB(reg_a, INDEXED_BYTE, 4+cycles); }
/* 0xA1 CMPA indexed */
static Uint32 op_cmpa_indexed(void) { OP_CMP(reg_a, INDEXED_BYTE, 4+cycles); }
/* 0xA2 SBCA indexed */
static Uint32 op_sbca_indexed(void) { OP_SBC(reg_a, INDEXED_BYTE, 4+cycles); }
/* 0xA3 SUBD indexed */
static Uint32 op_subd_indexed(void) { OP_SUB16(reg_d, INDEXED_WORD, 6+cycles); }
/* 0xA4 ANDA indexed */
static Uint32 op_anda_indexed(void) { OP_AND(reg_a, INDEXED_BYTE, 4+cycles); }
/* 0xA5 BITA indexed */
static Uint32 op_bita_indexed(void) { OP_BIT(reg_a, INDEXED_BYTE, 4+cycles); }
/* 0xA6 LDA indexed */
static Uint32 op_lda_indexed(void) { OP_LD(reg_a, INDEXED_BYTE, 4+cycles); }
/* 0xA7 STA indexed */
static Uint32 op_sta_indexed(void) { OP_ST(reg_a, EAD_INDEXED, 4+cycles); }
/* 0xA8 EORA indexed */
static Uint32 op_eora_indexed(void) { OP_EOR(reg_a, INDEXED_BYTE, 4+cycles); }
/* 0xA9 ADCA indexed */
static Uint32 op_adca_indexed(void) { OP_ADC(reg_a, INDEXED_BYTE, 4+cycles); }
/* 0xAA ORA indexed */
static Uint32 op_ora_indexed(void) { OP_OR(reg_a, INDEXED_BYTE, 4+cycles); }
/* 0xAB ADDA indexed */
static Uint32 op_adda_indexed(void) { OP_ADD(reg_a, INDEXED_BYTE, 4+cycles); }
/* 0xAC CMPX indexed */
static Uint32 op_cmpx_indexed(void) { OP_CMP16(reg_x, INDEXED_WORD, 6+cycles); }
/* 0xAD JSR indexed */
static Uint32 op_jsr_indexed(void) { OP_JSR(EAD_INDEXED, 7+cycles); }
/* 0xAE LDX indexed */
static Uint32 op_ldx_indexed(void) { OP_LD16(reg_x, INDEXED_WORD, 5+cycles); }
/* 0xAF STX indexed */
static Uint32 op_stx_indexed(void) { OP_ST16(reg_x, EAD_INDEXED, 5+cycles); }

/* ------------------------------------------------------------------------- */

/* 0xB0 SUBA extended */
static Uint32 op_suba_extended(void) { OP_SUB(reg_a, EXTENDED_BYTE, 5); }
/* 0xB1 CMPA extended */
static Uint32 op_cmpa_extended(void) { OP_CMP(reg_a, EXTENDED_BYTE, 5); }
/* 0xB2 SBCA extended */
static Uint32 op_sbca_extended(void) { OP_SBC(reg_a, EXTENDED_BYTE, 5); }
/* 0xB3 SUBD extended */
static Uint32 op_subd_extended(void) { OP_SUB16(reg_d, EXTENDED_WORD, 7); }
/* 0xB4 ANDA extended */
static Uint32 op_anda_extended(void) { OP_AND(reg_a, EXTENDED_BYTE, 5); }
/* 0xB5 BITA extended */
static Uint32 op_bita_extended(void) { OP_BIT(reg_a, EXTENDED_BYTE, 5); }
/* 0xB6 LDA extended */
static Uint32 op_lda_extended(void) { OP_LD(reg_a, EXTENDED_BYTE, 5); }
/* 0xB7 STA extended */
static Uint32 op_sta_extended(void) { OP_ST(reg_a, EAD_EXTENDED, 5); }
/* 0xB8 EORA extended */
static Uint32 op_eora_extended(void) { OP_EOR(reg_a, EXTENDED_BYTE, 5); }
/* 0xB9 ADCA extended */
static Uint32 op_adca_extended(void) { OP_ADC(reg_a, EXTENDED_BYTE, 5); }
/* 0xBA ORA extended */
static Uint32 op_ora_extended(void) { OP_OR(reg_a, EXTENDED_BYTE, 5); }
/* 0xBB ADDA extended */
static Uint32 op_adda_extended(void) { OP_ADD(reg_a, EXTENDED_BYTE, 5); }
/* 0xBC CMPX extended */
static Uint32 op_cmpx_extended(void) { OP_CMP16(reg_x, EXTENDED_WORD, 6+cycles); }
/* 0xBD JSR extended */
static Uint32 op_jsr_extended(void) { OP_JSR(EAD_EXTENDED, 8); }
/* 0xBE LDX extended */
static Uint32 op_ldx_extended(void) { OP_LD16(reg_x, EXTENDED_WORD, 6); }
/* 0xBF STX extended */
static Uint32 op_stx_extended(void) { OP_ST16(reg_x, EAD_EXTENDED, 6); }

/* ------------------------------------------------------------------------- */

/* 0xC0 SUBB immediate */
static Uint32 op_subb(void) { OP_SUB(reg_b, IMMEDIATE_BYTE, 2); }
/* 0xC1 CMPB immediate */
static Uint32 op_cmpb(void) { OP_CMP(reg_b, IMMEDIATE_BYTE, 2); }
/* 0xC2 SBCB immediate */
static Uint32 op_sbcb(void) { OP_SBC(reg_b, IMMEDIATE_BYTE, 2); }
/* 0xC3 ADDD immediate */
static Uint32 op_addd(void) { OP_ADD16(reg_d, IMMEDIATE_WORD, 4); }
/* 0xC4 ANDB immediate */
static Uint32 op_andb(void) { OP_AND(reg_b, IMMEDIATE_BYTE, 2); }
/* 0xC5 BITB immediate */
static Uint32 op_bitb(void) { OP_BIT(reg_b, IMMEDIATE_BYTE, 2); }
/* 0xC6 LDB immediate */
static Uint32 op_ldb(void) { OP_LD(reg_b, IMMEDIATE_BYTE, 2); }
/* 0xC8 EORB immediate */
static Uint32 op_eorb(void) { OP_EOR(reg_b, IMMEDIATE_BYTE, 2); }
/* 0xC9 ADCB immediate */
static Uint32 op_adcb(void) { OP_ADC(reg_b, IMMEDIATE_BYTE, 2); }
/* 0xCA ORB immediate */
static Uint32 op_orb(void) { OP_OR(reg_b, IMMEDIATE_BYTE, 2); }
/* 0xCB ADDB immediate */
static Uint32 op_addb(void) { OP_ADD(reg_b, IMMEDIATE_BYTE, 2); }
/* 0xCC LDD immediate */
static Uint32 op_ldd(void) { OP_LD16(reg_d, IMMEDIATE_WORD, 3); }
/* 0xCE LDU immediate */
static Uint32 op_ldu(void) { OP_LD16(reg_u, IMMEDIATE_WORD, 3); }

/* ------------------------------------------------------------------------- */

/* 0xD0 SUBB direct */
static Uint32 op_subb_direct(void) { OP_SUB(reg_b, DIRECT_BYTE, 4); }
/* 0xD1 CMPB direct */
static Uint32 op_cmpb_direct(void) { OP_CMP(reg_b, DIRECT_BYTE, 4); }
/* 0xD2 SBCB direct */
static Uint32 op_sbcb_direct(void) { OP_SBC(reg_b, DIRECT_BYTE, 4); }
/* 0xD3 ADDD direct */
static Uint32 op_addd_direct(void) { OP_ADD16(reg_d, DIRECT_WORD, 6); }
/* 0xD4 ANDB direct */
static Uint32 op_andb_direct(void) { OP_AND(reg_b, DIRECT_BYTE, 4); }
/* 0xD5 BITB direct */
static Uint32 op_bitb_direct(void) { OP_BIT(reg_b, DIRECT_BYTE, 4); }
/* 0xD6 LDB direct */
static Uint32 op_ldb_direct(void) { OP_LD(reg_b, DIRECT_BYTE, 4); }
/* 0xD7 STB direct */
static Uint32 op_stb_direct(void) { OP_ST(reg_b, EAD_DIRECT, 4); }
/* 0xD8 EORB direct */
static Uint32 op_eorb_direct(void) { OP_EOR(reg_b, DIRECT_BYTE, 4); }
/* 0xD9 ADCB direct */
static Uint32 op_adcb_direct(void) { OP_ADC(reg_b, DIRECT_BYTE, 4); }
/* 0xDA ORB direct */
static Uint32 op_orb_direct(void) { OP_OR(reg_b, DIRECT_BYTE, 4); }
/* 0xDB ADDB direct */
static Uint32 op_addb_direct(void) { OP_ADD(reg_b, DIRECT_BYTE, 4); }
/* 0xDC LDD direct */
static Uint32 op_ldd_direct(void) { OP_LD16(reg_d, DIRECT_WORD, 5); }
/* 0xDD STD direct */
static Uint32 op_std_direct(void) { OP_ST16(reg_d, EAD_DIRECT, 5); }
/* 0xDE LDU direct */
static Uint32 op_ldu_direct(void) { OP_LD16(reg_u, DIRECT_WORD, 5); }
/* 0xDF STU direct */
static Uint32 op_stu_direct(void) { OP_ST16(reg_u, EAD_DIRECT, 5); }

/* ------------------------------------------------------------------------- */

/* 0xE0 SUBB indexed */
static Uint32 op_subb_indexed(void) { OP_SUB(reg_b, INDEXED_BYTE, 4+cycles); }
/* 0xE1 CMPB indexed */
static Uint32 op_cmpb_indexed(void) { OP_CMP(reg_b, INDEXED_BYTE, 4+cycles); }
/* 0xE2 SBCB indexed */
static Uint32 op_sbcb_indexed(void) { OP_SBC(reg_b, INDEXED_BYTE, 4+cycles); }
/* 0xE3 ADDD indexed */
static Uint32 op_addd_indexed(void) { OP_ADD16(reg_d, INDEXED_WORD, 6+cycles); }
/* 0xE4 ANDB indexed */
static Uint32 op_andb_indexed(void) { OP_AND(reg_b, INDEXED_BYTE, 4+cycles); }
/* 0xE5 BITB indexed */
static Uint32 op_bitb_indexed(void) { OP_BIT(reg_b, INDEXED_BYTE, 4+cycles); }
/* 0xE6 LDB indexed */
static Uint32 op_ldb_indexed(void) { OP_LD(reg_b, INDEXED_BYTE, 4+cycles); }
/* 0xE7 STB indexed */
static Uint32 op_stb_indexed(void) { OP_ST(reg_b, EAD_INDEXED, 4+cycles); }
/* 0xE8 EORB indexed */
static Uint32 op_eorb_indexed(void) { OP_EOR(reg_b, INDEXED_BYTE, 4+cycles); }
/* 0xE9 ADCB indexed */
static Uint32 op_adcb_indexed(void) { OP_ADC(reg_b, INDEXED_BYTE, 4+cycles); }
/* 0xEA ORB indexed */
static Uint32 op_orb_indexed(void) { OP_OR(reg_b, INDEXED_BYTE, 4+cycles); }
/* 0xEB ADDB indexed */
static Uint32 op_addb_indexed(void) { OP_ADD(reg_b, INDEXED_BYTE, 4+cycles); }
/* 0xEC LDD indexed */
static Uint32 op_ldd_indexed(void) { OP_LD16(reg_d, INDEXED_WORD, 5+cycles); }
/* 0xED STD indexed */
static Uint32 op_std_indexed(void) { OP_ST16(reg_d, EAD_INDEXED, 5+cycles); }
/* 0xEE LDU indexed */
static Uint32 op_ldu_indexed(void) { OP_LD16(reg_u, INDEXED_WORD, 5+cycles); }
/* 0xEF STU indexed */
static Uint32 op_stu_indexed(void) { OP_ST16(reg_u, EAD_INDEXED, 5+cycles); }

/* ------------------------------------------------------------------------- */

/* 0xF0 SUBB extended */
static Uint32 op_subb_extended(void) { OP_SUB(reg_b, EXTENDED_BYTE, 4); }
/* 0xF1 CMPB extended */
static Uint32 op_cmpb_extended(void) { OP_CMP(reg_b, EXTENDED_BYTE, 4); }
/* 0xF2 SBCB extended */
static Uint32 op_sbcb_extended(void) { OP_SBC(reg_b, EXTENDED_BYTE, 4); }
/* 0xF3 ADDD extended */
static Uint32 op_addd_extended(void) { OP_ADD16(reg_d, EXTENDED_WORD, 6); }
/* 0xF4 ANDB extended */
static Uint32 op_andb_extended(void) { OP_AND(reg_b, EXTENDED_BYTE, 4); }
/* 0xF5 BITB extended */
static Uint32 op_bitb_extended(void) { OP_BIT(reg_b, EXTENDED_BYTE, 4); }
/* 0xF6 LDB extended */
static Uint32 op_ldb_extended(void) { OP_LD(reg_b, EXTENDED_BYTE, 4); }
/* 0xF7 STB extended */
static Uint32 op_stb_extended(void) { OP_ST(reg_b, EAD_EXTENDED, 5); }
/* 0xF8 EORB extended */
static Uint32 op_eorb_extended(void) { OP_EOR(reg_b, EXTENDED_BYTE, 4); }
/* 0xF9 ADCB extended */
static Uint32 op_adcb_extended(void) { OP_ADC(reg_b, EXTENDED_BYTE, 4); }
/* 0xFA ORB extended */
static Uint32 op_orb_extended(void) { OP_OR(reg_b, EXTENDED_BYTE, 4); }
/* 0xFB ADDB extended */
static Uint32 op_addb_extended(void) { OP_ADD(reg_b, EXTENDED_BYTE, 4); }
/* 0xFC LDD extended */
static Uint32 op_ldd_extended(void) { OP_LD16(reg_d, EXTENDED_WORD, 5); }
/* 0xFD STD extended */
static Uint32 op_std_extended(void) { OP_ST16(reg_d, EAD_EXTENDED, 5); }
/* 0xFE LDU extended */
static Uint32 op_ldu_extended(void) { OP_LD16(reg_u, EXTENDED_WORD, 5); }
/* 0xFF STU extended */
static Uint32 op_stu_extended(void) { OP_ST16(reg_u, EAD_EXTENDED, 5); }

/* ------------------------------------------------------------------------- */

/* 0x1021 LBRN relative */
static Uint32 op_lbrn(void) {
	LONG_RELATIVE;
	return 5;
}

/* 0x1022 LBHI relative */
static Uint32 op_lbhi(void) {
	LONG_RELATIVE;
	if (!(reg_cc & (CC_Z|CC_C))) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x1023 LBLS relative */
static Uint32 op_lbls(void) {
	LONG_RELATIVE;
	if (reg_cc & (CC_Z|CC_C)) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x1024 LBHS,BCC relative */
static Uint32 op_lbhs(void) {
	LONG_RELATIVE;
	if (!(reg_cc & CC_C)) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x1025 LBLO,BCS relative */
static Uint32 op_lblo(void) {
	LONG_RELATIVE;
	if (reg_cc & CC_C) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x1026 LBNE relative */
static Uint32 op_lbne(void) {
	LONG_RELATIVE;
	if (!(reg_cc & CC_Z)) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x1027 LBEQ relative */
static Uint32 op_lbeq(void) {
	LONG_RELATIVE;
	if (reg_cc & CC_Z) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x1028 LBVC relative */
static Uint32 op_lbvc(void) {
	LONG_RELATIVE;
	if (!(reg_cc & CC_V)) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x1029 LBVS relative */
static Uint32 op_lbvs(void) {
	LONG_RELATIVE;
	if (reg_cc & CC_V) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x102A LBPL relative */
static Uint32 op_lbpl(void) {
	LONG_RELATIVE;
	if (!(reg_cc & CC_N)) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x102B LBMI relative */
static Uint32 op_lbmi(void) {
	LONG_RELATIVE;
	if (reg_cc & CC_N) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x102C LBGE relative */
static Uint32 op_lbge(void) {
	LONG_RELATIVE;
	if (!N_EOR_V) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x102D LBLT relative */
static Uint32 op_lblt(void) {
	LONG_RELATIVE;
	if (N_EOR_V) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x102E LBGT relative */
static Uint32 op_lbgt(void) {
	LONG_RELATIVE;
	if (!(N_EOR_V || reg_cc & CC_Z)) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* 0x102F LBLE relative */
static Uint32 op_lble(void) {
	LONG_RELATIVE;
	if (N_EOR_V || reg_cc & CC_Z) {
		reg_pc += RA;
		return 6;
	}
	return 5;
}

/* ------------------------------------------------------------------------- */

/* 0x103F SWI2 inherent */
static Uint32 op_swi2(void) {
	reg_cc |= CC_E;
	PUSHWORD(reg_s, reg_pc);
	PUSHWORD(reg_s, reg_u);
	PUSHWORD(reg_s, reg_y);
	PUSHWORD(reg_s, reg_x);
	PUSHBYTE(reg_s, reg_dp);
	PUSHBYTE(reg_s, reg_b);
	PUSHBYTE(reg_s, reg_a);
	PUSHBYTE(reg_s, reg_cc);
	reg_pc = sam_read_word(0xfff4);
	return 20;
}

/* ------------------------------------------------------------------------- */

/* 0x1083 CMPD immediate */
static Uint32 op_cmpd(void) { OP_CMP16(reg_d, IMMEDIATE_WORD, 5); }
/* 0x108C CMPY immediate */
static Uint32 op_cmpy(void) { OP_CMP16(reg_y, IMMEDIATE_WORD, 5); }
/* 0x108E LDY immediate */
static Uint32 op_ldy(void) { OP_LD16(reg_y, IMMEDIATE_WORD, 4); }

/* ------------------------------------------------------------------------- */

/* 0x1093 CMPD direct */
static Uint32 op_cmpd_direct(void) { OP_CMP16(reg_d, DIRECT_WORD, 7); }
/* 0x109C CMPY direct */
static Uint32 op_cmpy_direct(void) { OP_CMP16(reg_y, DIRECT_WORD, 7); }
/* 0x109E LDY direct */
static Uint32 op_ldy_direct(void) { OP_LD16(reg_y, DIRECT_WORD, 6); }
/* 0x109F STY direct */
static Uint32 op_sty_direct(void) { OP_ST16(reg_y, EAD_DIRECT, 6); }

/* ------------------------------------------------------------------------- */

/* 0x10A3 CMPD indexed */
static Uint32 op_cmpd_indexed(void) { OP_CMP16(reg_d, INDEXED_WORD, 7+cycles); }
/* 0x10AC CMPY indexed */
static Uint32 op_cmpy_indexed(void) { OP_CMP16(reg_y, INDEXED_WORD, 7+cycles); }
/* 0x10AE LDY indexed */
static Uint32 op_ldy_indexed(void) { OP_LD16(reg_y, INDEXED_WORD, 6+cycles); }
/* 0x10AF STY indexed */
static Uint32 op_sty_indexed(void) { OP_ST16(reg_y, EAD_INDEXED, 6+cycles); }

/* ------------------------------------------------------------------------- */

/* 0x10B3 CMPD extended */
static Uint32 op_cmpd_extended(void) { OP_CMP16(reg_d, EXTENDED_WORD, 8); }
/* 0x10BC CMPY extended */
static Uint32 op_cmpy_extended(void) { OP_CMP16(reg_y, EXTENDED_WORD, 8); }
/* 0x10BE LDY extended */
static Uint32 op_ldy_extended(void) { OP_LD16(reg_y, EXTENDED_WORD, 7); }
/* 0x10BF STY extended */
static Uint32 op_sty_extended(void) { OP_ST16(reg_y, EAD_EXTENDED, 7); }

/* ------------------------------------------------------------------------- */

/* 0x10CE LDS immediate */
static Uint32 op_lds(void) { OP_LD16(reg_s, IMMEDIATE_WORD, 4); }
/* 0x10DE LDS direct */
static Uint32 op_lds_direct(void) { OP_LD16(reg_s, DIRECT_WORD, 6); }
/* 0x10DF STS direct */
static Uint32 op_sts_direct(void) { OP_ST16(reg_s, EAD_DIRECT, 6); }
/* 0x10EE LDS indexed */
static Uint32 op_lds_indexed(void) { OP_LD16(reg_s, INDEXED_WORD, 6+cycles); }
/* 0x10EF STS indexed */
static Uint32 op_sts_indexed(void) { OP_ST16(reg_s, EAD_INDEXED, 6+cycles); }
/* 0x10FE LDS extended */
static Uint32 op_lds_extended(void) { OP_LD16(reg_s, EXTENDED_WORD, 7); }
/* 0x10FF STS extended */
static Uint32 op_sts_extended(void) { OP_ST16(reg_s, EAD_EXTENDED, 7); }

/* ------------------------------------------------------------------------- */

/* 0x113F SWI3 inherent */
static Uint32 op_swi3(void) {
	reg_cc |= CC_E;
	PUSHWORD(reg_s, reg_pc);
	PUSHWORD(reg_s, reg_u);
	PUSHWORD(reg_s, reg_y);
	PUSHWORD(reg_s, reg_x);
	PUSHBYTE(reg_s, reg_dp);
	PUSHBYTE(reg_s, reg_b);
	PUSHBYTE(reg_s, reg_a);
	PUSHBYTE(reg_s, reg_cc);
	reg_pc = sam_read_word(0xfff2);
	return 20;
}

/* 0x1183 CMPU immediate */
static Uint32 op_cmpu(void) { OP_CMP16(reg_u, IMMEDIATE_WORD, 5); }
/* 0x118C CMPS immediate */
static Uint32 op_cmps(void) { OP_CMP16(reg_s, IMMEDIATE_WORD, 5); }
/* 0x1193 CMPU direct */
static Uint32 op_cmpu_direct(void) { OP_CMP16(reg_u, DIRECT_WORD, 7); }
/* 0x119C CMPS direct */
static Uint32 op_cmps_direct(void) { OP_CMP16(reg_s, DIRECT_WORD, 7); }
/* 0x11A3 CMPU indexed */
static Uint32 op_cmpu_indexed(void) { OP_CMP16(reg_u, INDEXED_WORD, 7+cycles); }
/* 0x11AC CMPS indexed */
static Uint32 op_cmps_indexed(void) { OP_CMP16(reg_s, INDEXED_WORD, 7+cycles); }
/* 0x11B3 CMPU extended */
static Uint32 op_cmpu_extended(void) { OP_CMP16(reg_u, EXTENDED_WORD, 8); }
/* 0x11BC CMPS extended */
static Uint32 op_cmps_extended(void) { OP_CMP16(reg_s, EXTENDED_WORD, 8); }

/* ------------------------------------------------------------------------- */

static Uint32 op_illegal(void) { return 1; }

/* ------------------------------------------------------------------------- */

static op_func m6809_page0[0x0100] = {
	op_neg_direct, op_negcom_direct, op_illegal, op_com_direct, op_lsr_direct, op_illegal, op_ror_direct, op_asr_direct, op_asl_direct, op_rol_direct, op_dec_direct, op_illegal, op_inc_direct, op_tst_direct, op_jmp_direct, op_clr_direct,
	op_page2, op_page3, op_nop, op_sync, op_illegal, op_illegal, op_lbra, op_lbsr, op_illegal, op_daa, op_orcc, op_illegal, op_andcc, op_sex, op_exg, op_tfr,
	op_bra, op_brn, op_bhi, op_bls, op_bhs, op_blo, op_bne, op_beq, op_bvc, op_bvs, op_bpl, op_bmi, op_bge, op_blt, op_bgt, op_ble,
	op_leax, op_leay, op_leas, op_leau, op_pshs, op_puls, op_pshu, op_pulu, op_illegal, op_rts, op_abx, op_rti, op_cwai, op_mul, op_illegal, op_swi,
	op_nega, op_illegal, op_illegal, op_coma, op_lsra, op_illegal, op_rora, op_asra, op_asla, op_rola, op_deca, op_illegal, op_inca, op_tsta, op_illegal, op_clra,
	op_negb, op_illegal, op_illegal, op_comb, op_lsrb, op_illegal, op_rorb, op_asrb, op_aslb, op_rolb, op_decb, op_illegal, op_incb, op_tstb, op_illegal, op_clrb,
	op_neg_indexed, op_illegal, op_illegal, op_com_indexed, op_lsr_indexed, op_illegal, op_ror_indexed, op_asr_indexed, op_asl_indexed, op_rol_indexed, op_dec_indexed, op_illegal, op_inc_indexed, op_tst_indexed, op_jmp_indexed, op_clr_indexed,
	op_neg_extended, op_illegal, op_illegal, op_com_extended, op_lsr_extended, op_illegal, op_ror_extended, op_asr_extended, op_asl_extended, op_rol_extended, op_dec_extended, op_illegal, op_inc_extended, op_tst_extended, op_jmp_extended, op_clr_extended,
	op_suba, op_cmpa, op_sbca, op_subd, op_anda, op_bita, op_lda, op_illegal, op_eora, op_adca, op_ora, op_adda, op_cmpx, op_bsr, op_ldx, op_illegal,
	op_suba_direct, op_cmpa_direct, op_sbca_direct, op_subd_direct, op_anda_direct, op_bita_direct, op_lda_direct, op_sta_direct, op_eora_direct, op_adca_direct, op_ora_direct, op_adda_direct, op_cmpx_direct, op_jsr_direct, op_ldx_direct, op_stx_direct,
	op_suba_indexed, op_cmpa_indexed, op_sbca_indexed, op_subd_indexed, op_anda_indexed, op_bita_indexed, op_lda_indexed, op_sta_indexed, op_eora_indexed, op_adca_indexed, op_ora_indexed, op_adda_indexed, op_cmpx_indexed, op_jsr_indexed, op_ldx_indexed, op_stx_indexed,
	op_suba_extended, op_cmpa_extended, op_sbca_extended, op_subd_extended, op_anda_extended, op_bita_extended, op_lda_extended, op_sta_extended, op_eora_extended, op_adca_extended, op_ora_extended, op_adda_extended, op_cmpx_extended, op_jsr_extended, op_ldx_extended, op_stx_extended,
	op_subb, op_cmpb, op_sbcb, op_addd, op_andb, op_bitb, op_ldb, op_illegal, op_eorb, op_adcb, op_orb, op_addb, op_ldd, op_illegal, op_ldu, op_illegal,
	op_subb_direct, op_cmpb_direct, op_sbcb_direct, op_addd_direct, op_andb_direct, op_bitb_direct, op_ldb_direct, op_stb_direct, op_eorb_direct, op_adcb_direct, op_orb_direct, op_addb_direct, op_ldd_direct, op_std_direct, op_ldu_direct, op_stu_direct,
	op_subb_indexed, op_cmpb_indexed, op_sbcb_indexed, op_addd_indexed, op_andb_indexed, op_bitb_indexed, op_ldb_indexed, op_stb_indexed, op_eorb_indexed, op_adcb_indexed, op_orb_indexed, op_addb_indexed, op_ldd_indexed, op_std_indexed, op_ldu_indexed, op_stu_indexed,
	op_subb_extended, op_cmpb_extended, op_sbcb_extended, op_addd_extended, op_andb_extended, op_bitb_extended, op_ldb_extended, op_stb_extended, op_eorb_extended, op_adcb_extended, op_orb_extended, op_addb_extended, op_ldd_extended, op_std_extended, op_ldu_extended, op_stu_extended
};

static op_func m6809_page2[0x0100] = {
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_lbrn, op_lbhi, op_lbls, op_lbhs, op_lblo, op_lbne, op_lbeq, op_lbvc, op_lbvs, op_lbpl, op_lbmi, op_lbge, op_lblt, op_lbgt, op_lble,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_swi2,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_cmpd, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_cmpy, op_illegal, op_ldy, op_illegal,
	op_illegal, op_illegal, op_illegal, op_cmpd_direct, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_cmpy_direct, op_illegal, op_ldy_direct, op_sty_direct,
	op_illegal, op_illegal, op_illegal, op_cmpd_indexed, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_cmpy_indexed, op_illegal, op_ldy_indexed, op_sty_indexed,
	op_illegal, op_illegal, op_illegal, op_cmpd_extended, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_cmpy_extended, op_illegal, op_ldy_extended, op_sty_extended,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_lds, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_lds_direct, op_sts_direct,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_lds_indexed, op_sts_indexed,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_lds_extended, op_sts_extended
};

static op_func m6809_page3[0x0100] = {
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_swi3,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_cmpu, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_cmps, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_cmpu_direct, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_cmps_direct, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_cmpu_indexed, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_cmps_indexed, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_cmpu_extended, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_cmps_extended, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal,
	op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal, op_illegal
};
