/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <SDL/SDL.h>

#define SOUND_SAMPLE_RATE 22050
#define SOUND_CHANNELS 1
#define SOUND_FORMAT AFMT_U8
#define SOUND_BUFFER_SETTING 0x0002000a
#define SOUND_FRAGMENT_SIZE 512
#define SOUND_BUFFER_SIZE (SOUND_FRAGMENT_SIZE + (SOUND_FRAGMENT_SIZE / 2))
#define SOUND_SUBCYCLES_PER_SAMPLE (((M6809_CPU_RATE*256)+128) / rate)
#define SOUND_SUBCYCLES_PER_FRAGMENT (((M6809_CPU_RATE * SOUND_FRAGMENT_SIZE * 256)+128) / rate)

static void sound_send_buffer(void);

static int sound_fd;
static int fragment;
static int channels;
static int format;
static int rate;

static Uint8 *buffer;
static Uint8 *wrptr;
static Uint8 lastsample;
static Uint32 start_subcycle;
static int sound_enable;
static Uint8 convbuf[SOUND_FRAGMENT_SIZE * 4];

int sound_init(void) {
	int tmp;
	fragment = SOUND_BUFFER_SETTING;
	channels = SOUND_CHANNELS - 1;
	format = SOUND_FORMAT;
	rate = SOUND_SAMPLE_RATE;
	sound_fd = open("/dev/dsp", O_WRONLY);
	if (sound_fd == -1) {
		LOG_ERROR0("Couldn't open audio device!\n");
		return 0;
	}
	tmp = fragment;
	if (ioctl(sound_fd, SNDCTL_DSP_SETFRAGMENT, &fragment) == -1 ||
			tmp != fragment) {
		LOG_ERROR0("Couldn't set audio buffer size/fragmentation limit.\n");
		return 0;
	}
	if (ioctl(sound_fd, SNDCTL_DSP_STEREO, &channels) == -1) {
		LOG_ERROR0("Couldn't set desired audio channels.\n");
		return 0;
	}
	channels++;
	if (ioctl(sound_fd, SNDCTL_DSP_SETFMT, &format) == -1) {
		LOG_ERROR0("Couldn't set desired audio format.\n");
		return 0;
	}
	if (ioctl(sound_fd, SNDCTL_DSP_SPEED, &rate) == -1) {
		LOG_ERROR0("Couldn't set desired audio rate.\n");
		return 0;
	}
	LOG_DEBUG3("Set up audio device at %dHz, %d channels.  Format = %02x.\n", rate, channels, format);
	if (format != SOUND_FORMAT || channels != SOUND_CHANNELS) {
		LOG_DEBUG0("Format needs conversion.\n");
	}
	PIA_1A.wrfunc[0] = sound_write_byte;
	buffer = (Uint8 *)malloc(SOUND_BUFFER_SIZE);
	wrptr = buffer;
	start_subcycle = 0;
	sound_enable = 0;
	return 1;
}

void sound_reset(void) {
}

void sound_write_byte(Uint8 octet) {
	Uint8 *new_wrptr = buffer + (((current_cycle<<8) - start_subcycle)/(Uint32)SOUND_SUBCYCLES_PER_SAMPLE);
	if (!(PIA_1B.control_register & 0x08) || PIA_0B.control_register & 0x08 || PIA_0A.control_register & 0x08)
		octet = lastsample; // 0x80;
	if (new_wrptr < buffer || new_wrptr >= buffer + SOUND_BUFFER_SIZE) {
		start_subcycle = current_cycle << 8;
		wrptr = buffer;
		lastsample = octet;
		return;
	}
	memset(wrptr, lastsample, new_wrptr - wrptr);
	wrptr = new_wrptr;
	if (new_wrptr >= buffer + SOUND_FRAGMENT_SIZE) {
		sound_send_buffer();
		wrptr -= SOUND_FRAGMENT_SIZE;
		start_subcycle -= SOUND_SUBCYCLES_PER_FRAGMENT;
		//memmove(buffer + SOUND_FRAGMENT_SIZE, buffer, SOUND_BUFFER_SIZE - SOUND_FRAGMENT_SIZE);
		memcpy(buffer, buffer + SOUND_FRAGMENT_SIZE, SOUND_BUFFER_SIZE - SOUND_FRAGMENT_SIZE);
	}
	lastsample = octet;
}

static void sound_send_buffer(void) {
	Uint8 *source = buffer;
	Uint8 *dest = convbuf;
	Uint8 tmp;
	int i;
	if (format == AFMT_U8 && channels == 1) {
		write(sound_fd, buffer, SOUND_FRAGMENT_SIZE);
		return;
	}
	if (format == AFMT_U8 && channels == 2) {
		for (i = SOUND_FRAGMENT_SIZE; i; i--) {
			tmp = *(source++);
			*(dest++) = tmp;
			*(dest++) = tmp;
		}
		write(sound_fd, convbuf, SOUND_FRAGMENT_SIZE * 2);
		return;
	}
	if (format == AFMT_S16_LE) {
		for (i = SOUND_FRAGMENT_SIZE; i; i--) {
			tmp = *(source++)^0x80;
			*(dest++) = 0;
			*(dest++) = tmp;
			if (channels == 2) {
				*(dest++) = 0;
				*(dest++) = tmp;
			}
		}
		write(sound_fd, convbuf, SOUND_FRAGMENT_SIZE * channels * 2);
		return;
	}
	if (format == AFMT_S16_BE) {
		for (i = SOUND_FRAGMENT_SIZE; i; i--) {
			tmp = *(source++)^0x80;
			*(dest++) = tmp;
			*(dest++) = 0;
			if (channels == 2) {
				*(dest++) = tmp;
				*(dest++) = 0;
			}
		}
		write(sound_fd, convbuf, SOUND_FRAGMENT_SIZE * channels * 2);
		return;
	}
}
